/*
 * plpmkboot - make fat fs plpbt.bin bootable
 *
 * Released under GPL v2
 * by Elmar Hanlhofer  http://www.plop.at
 *
 */


#define VERSION "0.1 03/Feb/2012"

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#ifdef WIN32
#include "windows.h"
#include "winioctl.h"
#endif

#include <stdlib.h>
#include <string.h>


#define TYPE_NOTFAT 0
#define TYPE_FAT12  1
#define TYPE_FAT16  2
#define TYPE_FAT32  3

char fat12ldr[] = {
#include "fat12ldr.h"
};

char fat16ldr[] = {
#include "fat16ldr.h"
};

char fat32ldr[] = {
#include "fat32ldr.h"
};



void PrintSuccess()
{
    printf ("\n");
    printf ("SUCCESS: The partition is now bootable with plpbt.bin.\n");
    printf ("Take care that the boot flag is set for the partition.\n\n");
}

#ifdef WIN32

HANDLE hDevice; // handle to the drive to be examined 

void OpenDrive (__const char *drive)
{
    char drivepath[256];
    
    sprintf (drivepath, "\\\\.\\%c:", drive[0]);

    hDevice = CreateFile(drivepath, // drive to open
			GENERIC_READ | GENERIC_WRITE, //
			FILE_SHARE_READ | FILE_SHARE_WRITE, // share mode
			NULL, // default security attributes
			OPEN_EXISTING, // disposition
			0, // file attributes
			NULL); // don't copy any file's attributes

    if (hDevice == INVALID_HANDLE_VALUE) // we can't open the drive
    {
	printf ("Unable to open %s\n", drive);
	exit (EXIT_FAILURE);    
    }
}
    
void ReadSectors (int start, int num, char *buffer)
{
    DWORD junk; // discard results
    LARGE_INTEGER pos;
    
    pos.QuadPart = start * 512;
    if (!SetFilePointerEx (hDevice, pos, NULL, FILE_BEGIN))
    {
	printf ("Cannot set file pointer to read\n");
	exit (EXIT_FAILURE);            
    }

    if (!ReadFile(hDevice, buffer, 512 * num, &junk, 0))
    {
	printf ("Unable to read sectors from the drive\n");
	exit (EXIT_FAILURE);        
    }
}

void WriteSectors (int start, int num, char *buffer)
{
    DWORD junk; // discard results
    LARGE_INTEGER pos;
    
    pos.QuadPart = start * 512;
    if (!SetFilePointerEx (hDevice, pos, NULL, FILE_BEGIN))
    {
	printf ("Cannot set file pointer to write\n");
	exit (EXIT_FAILURE);            
    }
    
    if (!WriteFile(hDevice, buffer, 512 * num, &junk, 0))
    {
	printf ("Unable to write to the sectors of the drive\n");
	exit (EXIT_FAILURE);        
    }
}

void CloseDrive()
{
    CloseHandle(hDevice); // we're done with the handle
}

#else

FILE *pDrive;

void OpenDrive (__const char *drive)
{

    pDrive = fopen (drive, "r+b");
    if (!pDrive)
    {
	printf ("Unable to open %s\n", drive);
	exit (EXIT_FAILURE);    
    }
}

void ReadSectors (int start, int num, char *buffer)
{
    fseek (pDrive, start * 512, SEEK_SET);
    
    if (ftell (pDrive) != start * 512)
    {
	printf ("Cannot set file pointer to read\n");
	exit (EXIT_FAILURE);            
    }
    
    if (!fread (buffer, num * 512, 1, pDrive))
    {
	printf ("Unable to read sectors from the drive\n");
	exit (EXIT_FAILURE);        
    }
}

void WriteSectors (int start, int num, char *buffer)
{
    fseek (pDrive, start * 512, SEEK_SET);
    
    if (ftell (pDrive) != start * 512)
    {
	printf ("Cannot set file pointer to write\n");
	exit (EXIT_FAILURE);            
    }
    
    if (!fwrite (buffer, num * 512, 1, pDrive))
    {
	printf ("Unable to write to the sectors of the drive\n");
	exit (EXIT_FAILURE);        
    }
}

void CloseDrive()
{
    fclose (pDrive);
}

int GetName (__const char *name)
{
    int from = strlen (name) - 1;
    for (int i = from; i >= 0; i--)
	if (name[i] == '/')
	    return i + 1;

    return 0;
}

#endif

FILE *BackupFile(__const char *drive, __const char *s)
{
    FILE *pf;
    int quit = 0;
    char name[40];
    int i = 0;
#ifndef WIN32
    int n = GetName (drive);
#endif    

    while (!quit)
    {
#ifdef WIN32
	if (i == 0)
	    sprintf (name, "sector%s_%c.bak", s, drive[0]);
	else
	    sprintf (name, "sector%s_%c_(%d).bak", s, drive[0], i);
#else	    
	if (i == 0)
	    sprintf (name, "sector%s_%s.bak", s, drive + n);
	else
	    sprintf (name, "sector%s_%s_(%d).bak", s, drive + n, i);
#endif	    
	pf = fopen (name, "r");
	if (pf)
	    fclose (pf);
	else
	    quit = 1;
	    
	i++;
	
    }

    pf = fopen (name, "wb");
    if (!pf)
    {
	printf ("Unable to create backup file %s\n", name);
	exit (EXIT_FAILURE);
    }
    
    printf ("Backup file: %s\n", name);
    return pf;
}

void InstallFat12 (char *s, __const char *drive)
{
    FILE *pf2;
    
    int i;
    
    printf ("FAT12 detected\n");

    pf2 = BackupFile(drive, "");
    

    fwrite (s, 512, 1, pf2);


    for (i = 0; i < 3; i++)
	s[i] = fat12ldr[i];

    for (i = 0x3E; i < 0x1FE; i++)
	s[i] = fat12ldr[i];


    WriteSectors (0, 1, s);

    fclose (pf2);
    
    PrintSuccess();

}


void InstallFat16 (char *s, __const char *drive)
{
    FILE *pf2;
    
    int i;
    
    printf ("FAT16 detected\n");

    pf2 = BackupFile(drive, "");
    

    fwrite (s, 512, 1, pf2);


    for (i = 0; i < 3; i++)
	s[i] = fat16ldr[i];

    for (i = 0x3E; i < 0x1FE; i++)
	s[i] = fat16ldr[i];


    WriteSectors (0, 1, s);

    fclose (pf2);
    
    PrintSuccess();

}


void InstallFat32 (char *s, __const char *drive)
{
    FILE *pf2;
    
    int i;
    char b[512 * 3];
    
    
    printf ("FAT32 detected\n");

    pf2 = BackupFile(drive, "s");
    

    unsigned int backup = (s[0x32] & 0xff) + ((s[0x33] & 0xff) << 8);
    
    printf ("Backup sector: %d\n", backup);
    
    fwrite (s, 512 * 3, 1, pf2);


    for (i = 0; i < 3; i++)
	s[i] = fat32ldr[i];

    for (i = 0x5A; i < 0x1FE; i++)
	s[i] = fat32ldr[i];


    // sector 3
    for (i = 512 * 2; i < 512 * 3; i++)
	s[i] = fat32ldr[i];

    WriteSectors (0, 3, s);


    // update backup boot sector

    
    ReadSectors (backup, 3, b);
    
    fwrite (b, 512 * 3, 1, pf2);

    for (i = 0; i < 3; i++)
	b[i] = fat32ldr[i];

    for (i = 0x5A; i < 0x1FE; i++)
	b[i] = fat32ldr[i];


    // sector 3
    for (i = 512 * 2; i < 512 * 3; i++)
	b[i] = fat32ldr[i];
    

    WriteSectors (backup, 3, b);

    fclose (pf2);
    
    PrintSuccess();

}


int GetFatType (__const char *buffer)
{
    if (strncasecmp (buffer + 0x36, "FAT12   ", 8) == 0)
	return TYPE_FAT12;

    if (strncasecmp (buffer + 0x36, "FAT16   ", 8) == 0)
	return TYPE_FAT16;
	
    if (strncasecmp (buffer + 0x52, "FAT32   ", 8) == 0)
	return TYPE_FAT32;
	
    return TYPE_NOTFAT;
}

void MakeBootable(__const char *drive)
{
    int type;
    char bootsectors[512 * 3];
    
    OpenDrive (drive);
    ReadSectors (0, 3, bootsectors);    
    
    type = GetFatType (bootsectors);

    switch (type)
    {
	case TYPE_FAT12:
	    InstallFat12 (bootsectors, drive);
	    break;

	case TYPE_FAT16:
	    InstallFat16 (bootsectors, drive);
	    break;
	
	case TYPE_FAT32:
	    InstallFat32 (bootsectors, drive);
	    break;
	    
	default:
	    printf ("Error: It's not FAT formatted\n");
    
    }
        
    CloseDrive();

}

void PrintHelp()
{
    printf ("plpmkboot " VERSION " by Elmar Hanlhofer http://www.plop.at\n\n");
    printf ("Make a FAT12/16/32 file system bootable to start plpbt.bin\n\n");
#ifdef WIN32
    printf ("plpmkboot [drive:]\n\n");
    printf ("Example: plpmkboot F:\n");
    printf ("Press any key...\n");
    getchar();
#else
    printf ("plpmkboot [device node]\n\n");
    printf ("Example: plpmkboot /dev/sdb1\n");
#endif
}

int main (int argc, char *argv[])
{
    if ((argc == 1) || (argc > 2) ||
	(strcmp (argv[1], "/?") == 0) ||
	(strcmp (argv[1], "-h") == 0) ||
	(strcmp (argv[1], "--help") == 0))
    {
	PrintHelp();
	exit (EXIT_FAILURE);    
    }

#ifdef WIN32
    if ((strlen (argv[1]) != 2) || (argv[1][1] != ':'))
    {
	printf ("Drive parameter error\n\n");
	
	PrintHelp();
	exit (EXIT_FAILURE);
    }
#endif

    MakeBootable(argv[1]);
    exit (EXIT_SUCCESS);
}
